# yapf: disable
import logging
import numpy as np
import torch
from torch import nn, optim
from tqdm import tqdm
from typing import List, Tuple, Union
from xrprimer.data_structure.camera import FisheyeCameraParameter

from xrmocap.data_structure.body_model import SMPLData
from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.model.body_model.builder import SMPL, build_body_model
from xrmocap.model.registrant.builder import SMPLify, build_registrant
from xrmocap.model.registrant.handler.builder import build_handler
from xrmocap.ops.projection.builder import PytorchProjector, build_projector
from xrmocap.transform.convention.keypoints_convention import convert_keypoints
from .base_optimizer import BaseOptimizer


# yapf: enable
class SMPLShapeAwareOptimizer(BaseOptimizer):

    def __init__(self,
                 smplify: Union[dict, SMPLify],
                 body_model: Union[dict, SMPL],
                 projector: Union[dict, PytorchProjector],
                 iteration=1,
                 refine_threshold=1,
                 kps2d_conf_threshold=0.97,
                 use_percep2d_optimizer: bool = False,
                 steps: int = 1000,
                 lr: float = 1e-3,
                 verbose: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """SMPL shape-aware optimizer.

        Args:
            smplify (Union[dict, SMPLify]): A SMPLify instance or its config.
            body_model (Union[dict, SMPL]):
                An instance of SMPL body_model or a config dict
                of SMPL body_model.
            projector (Union[dict, PytorchProjector]):
                An instance of PytorchProjector projector for points projection
                or a config dict of PytorchProjector projector.
            iteration (int, optional):
                The number of iterations optimized. Defaults to 1.
            refine_threshold (int, optional): Defaults to 1.
            kps2d_conf_threshold (float, optional):
                Threshold for 3D optimization using SMPL model.
                Defaults to 0.97.
            use_percep2d_optimizer (bool, optional):
                Whether to use 2D perception data for optimization.
                Defaults to False.
            steps (int, optional): Defaults to 1000.
            lr (float, optional): step size for Adam optimizer.
                Defaults to 1e-3.
            verbose (bool, optional): Whether to log info. Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        super().__init__(verbose=verbose, logger=logger)
        if isinstance(smplify, dict):
            smplify['logger'] = logger
            self.smplify = build_registrant(smplify)
        else:
            self.smplify = smplify
        if isinstance(body_model, dict):
            body_model['logger'] = logger
            self.body_model = build_body_model(body_model)
        else:
            self.body_model = body_model

        if isinstance(projector, dict):
            self.projector = build_projector(projector)
        else:
            self.projector = projector
        self.device = self.smplify.device
        self.refine_threshold = refine_threshold
        self.kps2d_conf_threshold = kps2d_conf_threshold
        self.iteration = iteration
        self.use_percep2d_optimizer = use_percep2d_optimizer
        self.steps = steps
        self.lr = lr

    def prepare_data(
        self, skps3d: np.ndarray, frame_idx: int, kp3d_idx: int,
        sperson: List[np.ndarray]
    ) -> Tuple[torch.Tensor, List[torch.Tensor], np.ndarray, list]:
        """Prepare data for one keypoint.

        Args:
            skps3d (np.ndarray): The keypoints3d from single person,
                in shape (n_kps3d, 3).
            frame_idx (int): The frame index.
            kp3d_idx (int): The keypoints 3d index.
            sperson (List[np.ndarray]): The matched person id.

        Returns:
            Tuple[torch.Tensor, List[torch.Tensor], np.ndarray, list]:
                A keypoint3d,  keypoints2d from different views, the
                selected views and keypoints2d conf from different views.
        """

        mview_n_person = len(np.where(~np.isnan(sperson))[0])
        reprojected_error = np.zeros(mview_n_person)
        mview_person_idx = 0
        kps2d_tensor = []
        mkps2d_conf_list = []
        for view, kps2d_idx in enumerate(sperson):
            if np.isnan(kps2d_idx):
                kps2d_tensor.append(torch.tensor(np.nan))
                continue
            f_person_idx = self.mview_person_id[view][frame_idx][int(
                sperson[view])]
            kp2d = self.kps2d_list[view][frame_idx, f_person_idx, kp3d_idx]
            mkps2d_conf_list.append(self.kps2d_conf_list[view][frame_idx,
                                                               f_person_idx,
                                                               kp3d_idx])

            kps2d_tensor.append(torch.tensor(kp2d, dtype=torch.float32))
            proj_kp2d = self.projector.project_single_point(
                torch.from_numpy(skps3d[kp3d_idx])).numpy()

            reprojected_error[mview_person_idx] += np.linalg.norm(
                proj_kp2d[view] - kp2d)
            mview_person_idx += 1
        selected_view = (reprojected_error - reprojected_error.mean()
                         ) / reprojected_error.std() < self.refine_threshold
        selected_view = selected_view & (
            reprojected_error < reprojected_error.mean())

        kp3d_tensor = torch.tensor(skps3d[kp3d_idx], dtype=torch.float32)
        return kp3d_tensor, kps2d_tensor, selected_view, mkps2d_conf_list

    def step(self, keypoints3d: Keypoints, keypoints2d: List[Keypoints],
             matched_list, keypoints_from_smpl: Keypoints) -> Keypoints:

        kps3d = keypoints3d.get_keypoints()[..., :3]
        kps3d_conf = keypoints3d.get_keypoints()[..., 3:4]
        kps3d_mask = keypoints3d.get_mask()
        n_kps3d = keypoints3d.get_keypoints_number()
        convention = keypoints3d.get_convention()

        kps3d_from_smpl = keypoints_from_smpl.get_keypoints()[..., :3]
        self.kps2d_list = []
        self.kps2d_conf_list = []
        for keypoints in keypoints2d:
            kps2d = keypoints.get_keypoints()[..., :2]
            kps2d_conf = keypoints.get_keypoints()[..., 2]
            self.kps2d_list.append(kps2d)
            self.kps2d_conf_list.append(kps2d_conf)
        for frame_idx, person in enumerate(tqdm(matched_list)):
            f_person_mask = np.where(
                np.sum(kps3d_mask[frame_idx], axis=-1) > 0)[0]
            for person_idx, sperson in enumerate(person):
                skps3d = kps3d[frame_idx, f_person_mask[person_idx]]
                for kp3d_idx in range(n_kps3d):
                    kp3d_tensor, kps2d_tensor, selected_view, mkps2d_conf_list\
                        = self.prepare_data(skps3d, frame_idx,
                                            kp3d_idx, sperson)

                    if self.use_percep2d_optimizer and selected_view.sum() > 0:
                        refined_kp3d = self.optimizer_based_percep2d(
                            kp3d_tensor, kps2d_tensor, selected_view)
                        kps3d[frame_idx, f_person_mask[person_idx],
                              kp3d_idx] = refined_kp3d.numpy().reshape(-1)
                    else:
                        high_level_conf = np.where(
                            np.array(mkps2d_conf_list) >
                            self.kps2d_conf_threshold)[0]
                        if len(high_level_conf) < 2:
                            kp3d_tensor_from_smpl = torch.tensor(
                                kps3d_from_smpl[frame_idx,
                                                f_person_mask[person_idx],
                                                kp3d_idx],
                                dtype=torch.float32)
                            refined_kp3d = self.optimizer_based_smpl(
                                kp3d_tensor, kp3d_tensor_from_smpl)
                            kps3d[frame_idx, f_person_mask[person_idx],
                                  kp3d_idx] = refined_kp3d.numpy().reshape(-1)

        kps3d = np.concatenate((kps3d, kps3d_conf), axis=-1)
        ret_keypoints = Keypoints(
            kps=kps3d, mask=kps3d_mask, convention=convention)
        return ret_keypoints

    def optimize_keypoints3d(self, keypoints3d: Keypoints,
                             keypoints2d: List[Keypoints],
                             mview_person_id: list, matched_list: list,
                             cam_params: List[FisheyeCameraParameter],
                             **kwargs: dict) -> Keypoints:
        """Forward function of keypoints3d optimizer.

        Args:
            keypoints3d (Keypoints): A keypoints3d Keypoints instance.
            keypoints2d (Keypoints): A list of keypoints2d Keypoints instance.
            mview_person_id (list):
                The kps2d index in the bbox2d threshold of different frames
                from different views.
            matched_list (list): The matched kps2d index from different views,
                np.nan indicates that no person is observed in this view.
            cam_params (List[FisheyeCameraParameter]): Camera parameter.
            kwargs:
                Redundant keyword arguments to be
                ignored.

        Returns:
            Keypoints: The optimized keypoints3d.
        """
        self.mview_person_id = mview_person_id
        self.projector.set_cameras(cam_params)
        for step in range(self.iteration):
            smpl_data_list = self.estimate_smpl(keypoints3d=keypoints3d)
            kps3d_list = []
            for smpl_data in smpl_data_list:
                body_model_kwargs = smpl_data.to_tensor_dict()
                body_model_output = self.body_model(**body_model_kwargs)
                model_joints = body_model_output['joints']
                kps3d_list.append(model_joints)
            kps3d = torch.stack(kps3d_list, dim=1).detach().numpy()
            kps3d_ones = np.ones_like(kps3d[..., 0:1])
            kps3d = np.concatenate((kps3d, kps3d_ones), axis=-1)
            keypoints3d_45 = Keypoints(kps=kps3d, convention='smpl_45')
            smpl_keypoints3d = convert_keypoints(
                keypoints3d_45, 'coco', approximate=True)
            kps3d_conf = np.ones_like(smpl_keypoints3d.get_keypoints()[...,
                                                                       0:1])
            smpl_keypoints3d.set_keypoints(
                np.concatenate(
                    (smpl_keypoints3d.get_keypoints()[..., :3], kps3d_conf),
                    axis=-1))

            keypoints3d = self.step(keypoints3d, keypoints2d, matched_list,
                                    smpl_keypoints3d)
        return keypoints3d

    def optimizer_based_percep2d(self, init_kps3d, mview_kps2d_tensor,
                                 selected_view) -> torch.Tensor:

        optim_kps3d = nn.Parameter(init_kps3d)
        optimizer = optim.Adam([optim_kps3d], lr=self.lr)
        last_loss = torch.tensor(0.)

        for step in range(self.steps):
            optimizer.zero_grad()
            loss = torch.tensor(0.)
            proj_kps2d = self.projector.project_single_point(optim_kps3d)
            for i, (is_selected, kps2d) in enumerate(
                    zip(selected_view, mview_kps2d_tensor)):
                if is_selected and not torch.isnan(kps2d).any():
                    loss += torch.norm(proj_kps2d[i] - kps2d)

            if torch.abs(last_loss - loss) < 1e-3:
                break
            else:
                last_loss = loss
            loss.backward()
            optimizer.step()
        return optim_kps3d.detach()

    def optimizer_based_smpl(self, init_kps3d,
                             kps3d_from_smpl) -> torch.Tensor:

        optim_kps3d = nn.Parameter(init_kps3d)
        optimizer = optim.Adam([optim_kps3d], lr=self.lr)
        last_loss = torch.tensor(0.)

        for step in range(self.steps):
            optimizer.zero_grad()
            loss = torch.linalg.norm(kps3d_from_smpl - optim_kps3d, ord=2)

            if torch.abs(last_loss - loss) < 1e-3:
                break
            else:
                last_loss = loss
            loss.backward()
            optimizer.step()
        return optim_kps3d.detach()

    def estimate_smpl(self, keypoints3d: Keypoints) -> SMPLData:
        """Estimate smpl parameters according to keypoints3d.

        Args:
            keypoints3d (Keypoints):
                A keypoints3d Keypoints instance, with only one person
                inside. This method will take the person at
                keypoints3d.get_keypoints()[:, 0, ...] to run smplify.

        Returns:
            SMPLData:
                SMPL data of the person.
        """
        self.logger.info('Estimating SMPL.')
        working_convention = 'smpl'
        n_frame = keypoints3d.get_frame_number()
        n_person = keypoints3d.get_person_number()
        keypoints3d = keypoints3d.to_tensor(device=self.device)
        person_mask = keypoints3d.get_mask()
        person_mask = torch.sum(person_mask, dim=2) > 0

        keypoints3d = convert_keypoints(
            keypoints=keypoints3d, dst=working_convention)
        kps3d_tensor = keypoints3d.get_keypoints()[:, :, :, :3].float()
        kps3d_conf = keypoints3d.get_mask()[:, :, ...]

        smpl_data_list = []
        for person in range(n_person):
            if person_mask[:, person].sum() == 0:
                continue
            global_orient = torch.zeros((n_frame, 3)).to(self.device)
            transl = torch.full((n_frame, 3), 1000.0).to(self.device)
            body_pose = torch.zeros((n_frame, 69)).to(self.device)
            betas = torch.zeros((n_frame, 10)).to(self.device)
            s_kps3d_tensor = kps3d_tensor[:, person][person_mask[:, person]]
            s_kps3d_conf = kps3d_conf[:, person][person_mask[:, person]]
            # build and run
            kp3d_mse_input = build_handler(
                dict(
                    type='Keypoint3dMSEInput',
                    keypoints3d=s_kps3d_tensor,
                    keypoints3d_conf=s_kps3d_conf,
                    keypoints3d_convention=working_convention,
                    handler_key='keypoints3d_mse'))
            kp3d_llen_input = build_handler(
                dict(
                    type='Keypoint3dLimbLenInput',
                    keypoints3d=s_kps3d_tensor,
                    keypoints3d_conf=s_kps3d_conf,
                    keypoints3d_convention=working_convention,
                    handler_key='keypoints3d_limb_len'))
            registrant_output = self.smplify(
                input_list=[kp3d_mse_input, kp3d_llen_input])

            smpl_data = SMPLData()
            global_orient[
                person_mask[:, person]] = registrant_output['global_orient']
            transl[person_mask[:, person]] = registrant_output['transl']
            body_pose[person_mask[:, person]] = registrant_output['body_pose']
            betas[person_mask[:, person]] = registrant_output['betas']
            output = {}
            output['global_orient'] = global_orient
            output['transl'] = transl
            output['body_pose'] = body_pose
            output['betas'] = betas
            smpl_data.from_param_dict(output)
            smpl_data_list.append(smpl_data)
        return smpl_data_list



import numpy as np
import torch
from typing import Union


def affine_transform_pts(
        pts: Union[list, torch.Tensor],
        t: Union[list, torch.Tensor]) -> Union[list, torch.Tensor]:
    """Affine transformation for points.

    Args:
        pts (Union[list, torch.Tensor]):
            Point(s) to be transferred.
            Nx2 or [batch_size, n_views, N, 2]
        t (Union[list, torch.Tensor]):
            Transformation. Nx2 or [batch_size, n_views, N, 2]

    Returns:
        pts_trans(Union[list, torch.Tensor]):
            Affine transformed point(s).
    """

    if not hasattr(pts, 'device'):
        if pts.ndim == 1:
            pts_trans = np.array([pts[0], pts[1], 1.]).T
            pts_trans = np.dot(t, pts_trans)
            return pts_trans[:2]
        else:
            raise NotImplementedError('Batch affine transform for  \
                point on CPU is not implemented')
    else:
        pts_homo = torch.cat(
            [pts, torch.ones(pts.shape[:-1] + (1, ), device=pts.device)],
            dim=-1)
        pts_trans = torch.matmul(pts_homo, t.transpose(2, 3))
        return pts_trans



import logging
import prettytable
from typing import List, TypeVar, Union
from xrprimer.utils.log_utils import get_logger

from xrmocap.model.registrant.handler.base_handler import BaseInput
from .smplify_base_hook import SMPLifyBaseHook

_SMPLify = TypeVar('_SMPLify')


class SMPLifyVerboseHook(SMPLifyBaseHook):
    """A hook for logging SMPLify information, controlled by logger level and
    SMPLify attributes."""

    def before_stage(self, registrant: _SMPLify, input_list: List[BaseInput],
                     optim_param: dict, stage_config: dict) -> None:
        """This hook will be triggered at the beginning of
        SMPLify.__optimize_stage__().

        Args:
            registrant (_SMPLify):
                The Registrant instance.
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            optim_param (dict):
                A dictionary that includes body model parameters,
                and optional attributes such as vertices and joints.
                Its keys can be found in
                registrant.__class__.OPTIM_PARAM.
            stage_config (dict):
                A dictionary that includes stage configurations.
        """
        if registrant.verbose:
            epoch_idx = stage_config['epoch_idx']
            stage_idx = stage_config['stage_idx']
            registrant.logger.info(f'epoch {epoch_idx}, stage {stage_idx}')

    def after_stage(self, registrant: _SMPLify, input_list: List[BaseInput],
                    optim_param: dict, stage_config: dict,
                    loss_dict: dict) -> None:
        """This hook will be triggered at the end of
        SMPLify.__optimize_stage__().

        Args:
            registrant (_SMPLify):
                The Registrant instance.
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            optim_param (dict):
                A dictionary that includes body model parameters,
                and optional attributes such as vertices and joints.
                Its keys can be found in
                registrant.__class__.OPTIM_PARAM.
            stage_config (dict):
                A dictionary that includes stage configurations.
            loss_dict (dict):
                A dict that contains all losses of the last evaluation.
        """
        if registrant.verbose:
            SMPLifyVerboseHook.log_losses(loss_dict, registrant.logger)

    def after_evaluate(self, registrant: _SMPLify, input_list: List[BaseInput],
                       optim_param: dict, loss_dict: dict) -> None:
        """This hook will be triggered at the end of SMPLify.evaluate().

        Args:
            registrant (_SMPLify):
                The Registrant instance.
            input_list (List[BaseInput]):
                Additional input for loss handlers. Each element is
                an instance of subclass of BaseInput.
            optim_param (dict):
                A dictionary that includes body model parameters,
                and optional attributes such as vertices and joints.
                Its keys can be found in
                registrant.__class__.OPTIM_PARAM.
            loss_dict (dict):
                A dict that contains all losses of this evaluation.
        """
        if registrant.verbose and registrant.info_level == 'step':
            SMPLifyVerboseHook.log_losses(loss_dict, registrant.logger)

    @staticmethod
    def log_losses(loss_dict: dict,
                   logger: Union[None, str, logging.Logger] = None) -> None:
        """Log the losses in loss_dict.

        Args:
            loss_dict (dict):
                A dict that contains all losses of this evaluation.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        logger = get_logger(logger)
        table = prettytable.PrettyTable()
        table.field_names = ['Loss name', 'Loss value']
        for key, value in loss_dict.items():
            if isinstance(value, float) or \
                    isinstance(value, int) or \
                    len(value.shape) == 0:
                table.add_row([key, f'{value:.6f}'])
            else:
                table.add_row([key, 'Not a scalar'])
        table = '\n' + table.get_string()
        logger.info(table)



# yapf: disable
from mmcv.utils import Registry

from .base_estimator import BaseEstimator
from .mperson_smpl_estimator import MultiPersonSMPLEstimator
from .mview_mperson_topdown_estimator import (
    MultiViewMultiPersonTopDownEstimator,
)
from .mview_sperson_smpl_estimator import MultiViewSinglePersonSMPLEstimator

# yapf: enable

ESTIMATORS = Registry('estimator')
ESTIMATORS.register_module(
    name='MultiViewSinglePersonSMPLEstimator',
    module=MultiViewSinglePersonSMPLEstimator)
ESTIMATORS.register_module(
    name='MultiPersonSMPLEstimator', module=MultiPersonSMPLEstimator)
ESTIMATORS.register_module(
    name='MultiViewMultiPersonTopDownEstimator',
    module=MultiViewMultiPersonTopDownEstimator)


def build_estimator(cfg) -> BaseEstimator:
    """Build estimator."""
    return ESTIMATORS.build(cfg)



from .body_model import SMPLData, SMPLXData, SMPLXDData
from .keypoints import Keypoints
from .limbs import Limbs
from .smc_reader import SMCReader

__all__ = [
    'Keypoints', 'Limbs', 'SMCReader', 'SMPLData', 'SMPLXDData', 'SMPLXData'
]



import logging
from typing import Union
from xrprimer.utils.log_utils import get_logger

from xrmocap.data_structure.keypoints import Keypoints


class BaseOptimizer:

    def __init__(self,
                 verbose: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Base class for keypoints3d optimizer.

        Args:
            verbose (bool, optional):
                Whether to log info.
                Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        self.verbose = verbose
        self.logger = get_logger(logger)

    def optimize_keypoints3d(self, keypoints3d: Keypoints,
                             **kwargs: dict) -> Keypoints:
        """Forward function of keypoints3d optimizer.

        Args:
            keypoints3d (Keypoints): Input keypoints3d.
        kwargs:
            Redundant keyword arguments to be
            ignored, including:
                mview_keypoints2d

        Returns:
            Keypoints: The optimized keypoints3d.
        """
        raise NotImplementedError



import json
import math
import numpy as np


class LimbInfo():

    def __init__(self, kps_convention) -> None:
        self.kps_convention = kps_convention
        with open('./weight/limb_info.json', 'r') as f:
            self.info_dict = json.load(f)[self.kps_convention]

    def get_kps_number(self):
        """get keypoints number."""
        return self.info_dict['n_kps']

    def get_paf_number(self):
        """get paf number."""
        return self.info_dict['n_pafs']

    def get_shape_size(self):
        """get the prior shape number."""
        return self.info_dict['shape_size']

    def get_kps_parent(self):
        """get keypoints parent list."""
        return self.info_dict['kps_parent']

    def get_shape_blend(self):
        """get shape blend."""
        return self.info_dict['shape_blend']

    def get_kps_prior(self):
        """get prior keypoints."""
        return self.info_dict['m_kps']

    def get_hierarchy_map(self):
        """get hierarchy map for keypoints."""
        return self.info_dict['hierarchy_map']

    def get_paf_dict(self):
        """get paf dict."""
        return self.info_dict['paf_dict']


def welsch(c, x):
    x = x / c
    return 1 - math.exp(-x * x / 2)


def line2linedist(pa, raya, pb, rayb):
    if abs(np.vdot(raya, rayb)) < 1e-5:
        return point2linedist(pa, pb, raya)
    else:
        ve = np.cross(raya, rayb)
        ve = ve / np.linalg.norm(ve)
        ve = abs(np.vdot((pa - pb), ve))
        return ve


def point2linedist(pa, pb, ray):
    ve = np.cross(pa - pb, ray)
    return np.linalg.norm(ve)


def skew(vec):
    m_skew = np.zeros((3, 3), dtype=np.float32)
    m_skew = np.array(
        [0, -vec[2], vec[1], vec[2], 0, -vec[0], -vec[1], vec[0], 0],
        dtype=np.float32).reshape((3, 3))
    return m_skew


def rodrigues(vec):
    theta = np.linalg.norm(vec)
    identity = np.identity(3, dtype=np.float32)
    if abs(theta) < 1e-5:
        return identity
    else:
        c = np.cos(theta)
        s = np.sin(theta)
        r = vec / theta
        return c * identity + np.matmul((1 - c) * r.reshape(
            (-1, 1)), r.reshape((1, -1))) + s * skew(r)


def rodrigues_jacobi(vec):
    theta = np.linalg.norm(vec)
    d_skew = np.zeros((3, 9), dtype=np.float32)
    d_skew[0, 5] = d_skew[1, 6] = d_skew[2, 1] = -1
    d_skew[0, 7] = d_skew[1, 2] = d_skew[2, 3] = 1
    if abs(theta) < 1e-5:
        return -d_skew
    else:
        c = np.cos(theta)
        s = np.sin(theta)
        c1 = 1 - c
        itheta = 1 / theta
        r = vec / theta
        rrt = np.matmul(r.reshape((-1, 1)), r.reshape((1, -1)))
        m_skew = skew(r)
        identity = np.identity(3, dtype=np.float32)
        drrt = np.array([
            r[0] + r[0], r[1], r[2], r[1], 0, 0, r[2], 0, 0, 0, r[0], 0, r[0],
            r[1] + r[1], r[2], 0, r[2], 0, 0, 0, r[0], 0, 0, r[1], r[0], r[1],
            r[2] + r[2]
        ],
                        dtype=np.float32).reshape((3, 9))
        jaocbi = np.zeros((3, 9), dtype=np.float32)
        a = np.zeros((5, 1), dtype=np.float32)
        for i in range(3):
            a = np.array([
                -s * r[i], (s - 2 * c1 * itheta) * r[i], c1 * itheta,
                (c - s * itheta) * r[i], s * itheta
            ],
                         dtype=np.float32).reshape((5, 1))
            for j in range(3):
                for k in range(3):

                    jaocbi[i, k + k + k + j] = (
                        a[0] * identity[j, k] + a[1] * rrt[j, k] +
                        a[2] * drrt[i, j + j + j + k] + a[3] * m_skew[j, k] +
                        a[4] * d_skew[i, j + j + j + k])
        return jaocbi



# yapf: disable
import numpy
import torch
from pytorch3d.transforms import (
    axis_angle_to_matrix, matrix_to_euler_angles, matrix_to_rotation_6d,
)
from typing import Union

from xrmocap.transform.convention.joints_convention.standard_joint_angles import (  # noqa:E501
    TRANSFORMATION_AA_TO_SJA, TRANSFORMATION_SJA_TO_AA,
)

# yapf: enable


class Compose:

    def __init__(self, transforms: list):
        """Composes several transforms together. This transform does not
        support torchscript.

        Args:
            transforms (list): (list of transform functions)
        """
        self.transforms = transforms

    def __call__(self,
                 rotation: Union[torch.Tensor, numpy.ndarray],
                 convention: str = 'xyz',
                 **kwargs):
        convention = convention.lower()
        if not (set(convention) == set('xyz') and len(convention) == 3):
            raise ValueError(f'Invalid convention {convention}.')
        if isinstance(rotation, numpy.ndarray):
            data_type = 'numpy'
            rotation = torch.FloatTensor(rotation)
        elif isinstance(rotation, torch.Tensor):
            data_type = 'tensor'
        else:
            raise TypeError(
                'Type of rotation should be torch.Tensor or numpy.ndarray')
        for t in self.transforms:
            if 'convention' in t.__code__.co_varnames:
                rotation = t(rotation, convention.upper(), **kwargs)
            else:
                rotation = t(rotation, **kwargs)
        if data_type == 'numpy':
            rotation = rotation.detach().cpu().numpy()
        return rotation


def aa_to_sja(
    axis_angle: Union[torch.Tensor, numpy.ndarray],
    R_t: Union[torch.Tensor, numpy.ndarray] = TRANSFORMATION_AA_TO_SJA,
    R_t_inv: Union[torch.Tensor, numpy.ndarray] = TRANSFORMATION_SJA_TO_AA
) -> Union[torch.Tensor, numpy.ndarray]:
    """Convert axis-angles to standard joint angles.

    Args:
        axis_angle (Union[torch.Tensor, numpy.ndarray]): input shape
                should be (..., 21, 3), ndim of input is unlimited.
        R_t (Union[torch.Tensor, numpy.ndarray]): input shape
                should be (..., 21, 3, 3). Transformation matrices from
                original axis-angle coordinate system to
                standard joint angle coordinate system,
                ndim of input is unlimited.
        R_t_inv (Union[torch.Tensor, numpy.ndarray]): input shape
                should be (..., 21, 3, 3). Transformation matrices from
                standard joint angle coordinate system to
                original axis-angle coordinate system,
                ndim of input is unlimited.

    Returns:
        Union[torch.Tensor, numpy.ndarray]: shape would be (..., 3).
    """

    def _aa_to_sja(aa, R_t, R_t_inv):
        R_aa = axis_angle_to_matrix(aa)
        R_sja = R_t @ R_aa @ R_t_inv
        sja = matrix_to_euler_angles(R_sja, convention='XYZ')
        return sja

    if axis_angle.shape[-2:] != (21, 3):
        raise ValueError(
            f'Invalid input axis angles shape f{axis_angle.shape}.')
    if R_t.shape[-3:] != (21, 3, 3):
        raise ValueError(f'Invalid input R_t shape f{R_t.shape}.')
    if R_t_inv.shape[-3:] != (21, 3, 3):
        raise ValueError(f'Invalid input R_t_inv shape f{R_t.shape}.')
    t = Compose([_aa_to_sja])
    return t(axis_angle, R_t=R_t, R_t_inv=R_t_inv)


def aa_to_rot6d(
    axis_angle: Union[torch.Tensor, numpy.ndarray]
) -> Union[torch.Tensor, numpy.ndarray]:
    """Convert axis angles to rotation 6d representations.

    Args:
        axis_angle (Union[torch.Tensor, numpy.ndarray]): input shape
                should be (..., 3). ndim of input is unlimited.

    Returns:
        Union[torch.Tensor, numpy.ndarray]: shape would be (..., 6).

    [1] Zhou, Y., Barnes, C., Lu, J., Yang, J., & Li, H.
    On the Continuity of Rotation Representations in Neural Networks.
    IEEE Conference on Computer Vision and Pattern Recognition, 2019.
    Retrieved from http://arxiv.org/abs/1812.07035
    """
    if axis_angle.shape[-1] != 3:
        raise ValueError(f'Invalid input axis_angle f{axis_angle.shape}.')
    t = Compose([axis_angle_to_matrix, matrix_to_rotation_6d])
    return t(axis_angle)



# yapf: disable
import logging
import numpy as np
import torch
from tqdm import tqdm
from typing import List, Tuple, Union, overload
from xrprimer.data_structure.camera import FisheyeCameraParameter

from xrmocap.data_structure.body_model import SMPLData
from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.human_perception.builder import (
    MMdetDetector, MMposeTopDownEstimator, build_detector,
)
from xrmocap.io.image import (
    get_n_frame_from_mview_src, load_clip_from_mview_src,
)
from xrmocap.model.registrant.builder import SMPLify
from xrmocap.ops.top_down_association.builder import (
    MvposeAssociator, build_top_down_associator,
)
from xrmocap.ops.triangulation.builder import (
    BaseTriangulator, build_triangulator,
)
from xrmocap.ops.triangulation.point_selection.builder import (
    BaseSelector, build_point_selector,
)
from xrmocap.transform.convention.keypoints_convention import (
    convert_keypoints, get_keypoint_num,
)
from xrmocap.transform.keypoints3d.optim.builder import BaseOptimizer
from .mperson_smpl_estimator import MultiPersonSMPLEstimator

# yapf: enable


class MultiViewMultiPersonTopDownEstimator(MultiPersonSMPLEstimator):
    """Api for estimating keypoints3d and smpl in a multi-view multi-person
    scene, using optimization-based top-down method."""

    def __init__(self,
                 bbox_thr: float,
                 work_dir: str,
                 bbox_detector: Union[dict, MMdetDetector],
                 kps2d_estimator: Union[dict, MMposeTopDownEstimator],
                 associator: Union[dict, MvposeAssociator],
                 smplify: Union[dict, SMPLify],
                 triangulator: Union[dict, BaseTriangulator],
                 point_selectors: List[Union[dict, BaseSelector, None]] = None,
                 kps3d_optimizers: Union[List[Union[BaseOptimizer, dict]],
                                         None] = None,
                 pred_kps3d_convention: str = 'coco',
                 load_batch_size: int = 10,
                 verbose: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Initialization of the class.

        Args:
            bbox_thr (float):
                The threshold of the bbox2d.
            work_dir (str):
                Path to the folder for running the api.
                No file in work_dir will be modified or added by
                MultiViewMultiPersonTopDownEstimator.
            bbox_detector (Union[dict, MMdetDetector]):
                A human bbox_detector or its config.
            kps2d_estimator (Union[dict, MMposeTopDownEstimator]):
                A top-down kps2d estimator or its config.
            associator (Union[dict, MvposeAssociator]):
                A MvposeAssociator instance or its config.
            smplify (Union[dict, SMPLify]):
                A SMPLify instance or its config.
            triangulator (Union[dict, BaseTriangulator]):
                A triangulator or its config.
            point_selectors (List[Union[dict, BaseSelector, None]], optional):
                A point selector or its config. If it's given, points
                will be selected before triangulation.
                Defaults to None.
            kps3d_optimizers (Union[
                    List[Union[BaseOptimizer, dict]], None], optional):
                A list of keypoints3d optimizers or their configs. If given,
                keypoints3d will be optimized by the cascaded final optimizers.
                Defaults to None.
            pred_kps3d_convention (str, optional): Defaults to 'coco'.
            load_batch_size (int, optional):
                How many frames are loaded at the same time. Defaults to 10.
            verbose (bool, optional):
                Whether to print(logger.info) information during estimating.
                Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        super().__init__(
            work_dir=work_dir,
            smplify=smplify,
            kps3d_optimizers=kps3d_optimizers,
            verbose=verbose,
            logger=logger)
        self.bbox_thr = bbox_thr
        self.load_batch_size = load_batch_size
        self.pred_kps3d_convention = pred_kps3d_convention

        if isinstance(bbox_detector, dict):
            bbox_detector['logger'] = logger
            self.bbox_detector = build_detector(bbox_detector)
        else:
            self.bbox_detector = bbox_detector

        if isinstance(kps2d_estimator, dict):
            kps2d_estimator['logger'] = logger
            self.kps2d_estimator = build_detector(kps2d_estimator)
        else:
            self.kps2d_estimator = kps2d_estimator

        if isinstance(associator, dict):
            associator['logger'] = logger
            self.associator = build_top_down_associator(associator)
        else:
            self.associator = associator

        if isinstance(triangulator, dict):
            triangulator['logger'] = logger
            self.triangulator = build_triangulator(triangulator)
        else:
            self.triangulator = triangulator

        if point_selectors is None:
            self.point_selectors = None
        else:
            self.point_selectors = []
            for selector in point_selectors:
                if isinstance(selector, dict):
                    selector['logger'] = logger
                    selector = build_point_selector(selector)
                self.point_selectors.append(selector)

    @overload
    def run(
        self, img_arr: np.ndarray, cam_param: List[FisheyeCameraParameter]
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        ...

    @overload
    def run(
        self, img_paths: List[List[str]],
        cam_param: List[FisheyeCameraParameter]
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        ...

    @overload
    def run(
        self, video_path: List[str], cam_param: List[FisheyeCameraParameter]
    ) -> Tuple[List[Keypoints], Keypoints, SMPLData]:
        ...

    def run(
        self,
        cam_param: List[FisheyeCameraParameter],
        img_arr: Union[None, np.ndarray] = None,
        img_paths: Union[None, List[List[str]]] = None,
        video_paths: Union[None, List[str]] = None,
    ) -> Tuple[Keypoints, List[SMPLData]]:
        """Run mutli-view multi-person topdown estimator once. run() needs one
        images input among [img_arr, img_paths, video_paths].

        Args:
            cam_param (List[FisheyeCameraParameter]):
                A list of FisheyeCameraParameter instances.
            img_arr (Union[None, np.ndarray], optional):
                A multi-view image array, in shape
                [n_view, n_frame, h, w, c]. Defaults to None.
            img_paths (Union[None, List[List[str]]], optional):
                A nested list of image paths, in shape
                [n_view, n_frame]. Defaults to None.
            video_paths (Union[None, List[str]], optional):
                A list of video paths, each is a view.
                Defaults to None.

        Returns:
            Tuple[Keypoints, List[SMPLData]]:
                A keypoints3d, a list of SMPLData.
        """
        input_list = [img_arr, img_paths, video_paths]
        input_count = 0
        for input_instance in input_list:
            if input_instance is not None:
                input_count += 1
        if input_count > 1:
            self.logger.error('Redundant input!\n' +
                              'Please offer only one among' +
                              ' img_arr, img_paths and video_paths.')
            raise ValueError
        elif input_count < 1:
            self.logger.error('No image input has been found!\n' +
                              'img_arr, img_paths and video_paths are None.')
            raise ValueError
        n_frame = get_n_frame_from_mview_src(img_arr, img_paths, video_paths,
                                             self.logger)
        self.associator.set_cameras(cam_param)
        self.triangulator.set_cameras(cam_param)
        if self.point_selectors is not None:
            for selector in self.point_selectors:
                if hasattr(selector, 'triangulator'):
                    selector.triangulator.set_cameras(cam_param)
        n_view = len(cam_param)
        n_kps = get_keypoint_num(convention=self.pred_kps3d_convention)
        pred_kps3d = np.zeros((n_frame, 1, n_kps, 4))
        association_results = [[] for _ in range(n_frame)]
        selected_keypoints2d = []
        max_identity = 0
        for start_idx in tqdm(
                range(0, n_frame, self.load_batch_size),
                disable=not self.verbose):
            end_idx = min(n_frame, start_idx + self.load_batch_size)
            mview_batch_arr = load_clip_from_mview_src(
                start_idx=start_idx,
                end_idx=end_idx,
                img_arr=img_arr,
                img_paths=img_paths,
                video_paths=video_paths,
                logger=self.logger)
            # Estimate bbox2d and keypoints2d
            bbox2d_list, keypoints2d_list = self.estimate_perception2d(
                mview_batch_arr)
            for sub_fidx in range(end_idx - start_idx):
                sframe_bbox2d_list = []
                sframe_keypoints2d_list = []
                for view_idx in range(n_view):
                    sview_kps2d_idx = []
                    for idx, bbox2d in enumerate(
                            bbox2d_list[view_idx][sub_fidx]):
                        if bbox2d[-1] > self.bbox_thr:
                            sview_kps2d_idx.append(idx)
                    sview_kps2d_idx = np.array(sview_kps2d_idx)
                    if len(sview_kps2d_idx) > 0:
                        sframe_bbox2d_list.append(
                            torch.tensor(bbox2d_list[view_idx][sub_fidx])
                            [sview_kps2d_idx])
                        mframe_keypoints2d = keypoints2d_list[view_idx]
                        keypoints2d = Keypoints(
                            kps=mframe_keypoints2d.get_keypoints()[
                                sub_fidx:sub_fidx + 1, sview_kps2d_idx],
                            mask=mframe_keypoints2d.get_mask()[
                                sub_fidx:sub_fidx + 1, sview_kps2d_idx],
                            convention=mframe_keypoints2d.get_convention(),
                            logger=self.logger)
                        if keypoints2d.get_convention(
                        ) != self.pred_kps3d_convention:
                            keypoints2d = convert_keypoints(
                                keypoints=keypoints2d,
                                dst=self.pred_kps3d_convention,
                                approximate=True)
                        sframe_keypoints2d_list.append(keypoints2d)
                    else:
                        sframe_bbox2d_list.append(torch.tensor([]))
                        sframe_keypoints2d_list.append(
                            Keypoints(
                                kps=np.zeros((1, 1, n_kps, 3)),
                                mask=np.zeros((1, 1, n_kps)),
                                convention=self.pred_kps3d_convention))
                # Establish cross-frame and cross-person associations
                sframe_association_results, predict_keypoints3d, identities = \
                    self.associator.associate_frame(
                        # Dimension definition varies between
                        # cv2 images and tensor images.
                        mview_img_arr=mview_batch_arr[:, sub_fidx].transpose(
                            0, 3, 1, 2),
                        mview_bbox2d=sframe_bbox2d_list,
                        mview_keypoints2d=sframe_keypoints2d_list,
                        affinity_type='geometry_mean'
                    )
                for p_idx in range(len(sframe_association_results)):
                    # Triangulation, one associated person per time
                    identity = identities[p_idx]
                    associate_idxs = sframe_association_results[p_idx]
                    tri_kps2d = np.zeros((n_view, n_kps, 3))
                    tri_mask = np.zeros((n_view, n_kps, 1))
                    for view_idx in range(n_view):
                        kps2d_idx = associate_idxs[view_idx]
                        if not np.isnan(kps2d_idx):
                            tri_kps2d[view_idx] = sframe_keypoints2d_list[
                                view_idx].get_keypoints()[0, int(kps2d_idx)]
                            tri_mask[view_idx, :, 0] = sframe_keypoints2d_list[
                                view_idx].get_mask()[0, int(kps2d_idx)]
                    if self.point_selectors is not None:
                        for selector in self.point_selectors:
                            tri_mask = selector.get_selection_mask(
                                points=tri_kps2d, init_points_mask=tri_mask)
                    kps3d = self.triangulator.triangulate(tri_kps2d, tri_mask)
                    if identity > max_identity:
                        n_identity = identity - max_identity
                        pred_kps3d = np.concatenate(
                            (pred_kps3d,
                             np.zeros((n_frame, n_identity, n_kps, 4))),
                            axis=1)
                        max_identity = identity
                    pred_kps3d[start_idx + sub_fidx,
                               identity] = np.concatenate(
                                   (kps3d, np.ones_like(kps3d[:, 0:1])),
                                   axis=-1)
                for identity in sorted(identities):
                    index = identities.index(identity)
                    association_results[start_idx + sub_fidx].append(
                        sframe_association_results[index])
                selected_keypoints2d.append(sframe_keypoints2d_list)
        # Convert array to keypoints instance
        pred_keypoints3d = Keypoints(
            dtype='numpy',
            kps=pred_kps3d,
            mask=pred_kps3d[..., -1] > 0,
            convention=self.pred_kps3d_convention,
            logger=self.logger)
        # Save keypoints2d
        selected_keypoints2d_list = []
        mview_person_id = [[] for _ in range(n_view)]
        for view_idx in range(n_view):
            pred_kps2d = np.zeros((n_frame, 1, n_kps, 3))
            max_n_kps2d = 1
            for frame_idx, sframe_keypoints2d in enumerate(
                    selected_keypoints2d):
                kps2d = sframe_keypoints2d[view_idx].get_keypoints()
                n_kps2d = sframe_keypoints2d[view_idx].get_person_number()
                if n_kps2d > max_n_kps2d:
                    pred_kps2d = np.concatenate(
                        (pred_kps2d,
                         np.zeros(
                             (n_frame, (n_kps2d - max_n_kps2d), n_kps, 3))),
                        axis=1)
                    max_n_kps2d = n_kps2d
                pred_kps2d[frame_idx, :n_kps2d] = kps2d[0]
                mview_person_id[view_idx].append(np.array(range(n_kps2d)))
            selected_keypoints2d_list.append(
                Keypoints(
                    kps=pred_kps2d,
                    mask=pred_kps2d[..., -1] > 0,
                    convention=self.pred_kps3d_convention))

        optim_kwargs = dict(
            keypoints2d=selected_keypoints2d_list,
            mview_person_id=mview_person_id,
            matched_list=association_results,
            cam_params=cam_param)
        # Optimizing keypoints3d
        pred_keypoints3d = self.optimize_keypoints3d(pred_keypoints3d,
                                                     **optim_kwargs)
        # Fitting SMPL model
        smpl_data_list = self.estimate_smpl(keypoints3d=pred_keypoints3d)

        return pred_keypoints3d, smpl_data_list

    def estimate_perception2d(
        self, img_arr: Union[None, np.ndarray]
    ) -> Tuple[List[np.ndarray], List[Keypoints]]:
        """Estimate bbox2d and keypoints2d.

        Args:
            img_arr (Union[None, np.ndarray], optional):
                A multi-view image array, in shape
                [n_view, n_frame, h, w, c]. Defaults to None.
        Returns:
            Tuple[List[np.ndarray], List[Keypoints]]:
                A list of bbox2d, and a list of keypoints2d instances.
        """
        mview_bbox2d_list = []
        mview_keypoints2d_list = []
        for view_index in range(img_arr.shape[0]):
            bbox2d_list = self.bbox_detector.infer_array(
                image_array=img_arr[view_index],
                disable_tqdm=(not self.verbose),
                multi_person=True)
            kps2d_list, _, bbox2d_list = self.kps2d_estimator.infer_array(
                image_array=img_arr[view_index],
                bbox_list=bbox2d_list,
                disable_tqdm=(not self.verbose))
            keypoints2d = self.kps2d_estimator.get_keypoints_from_result(
                kps2d_list)
            mview_bbox2d_list.append(bbox2d_list)
            mview_keypoints2d_list.append(keypoints2d)
        return mview_bbox2d_list, mview_keypoints2d_list



# yapf: disable
from .base_estimator import BaseEstimator
from .mperson_smpl_estimator import MultiPersonSMPLEstimator
from .mview_mperson_end2end_estimator import (
    MultiViewMultiPersonEnd2EndEstimator,
)
from .mview_mperson_topdown_estimator import (
    MultiViewMultiPersonTopDownEstimator,
)
from .mview_sperson_smpl_estimator import MultiViewSinglePersonSMPLEstimator

# yapf: enable

__all__ = [
    'BaseEstimator', 'MultiPersonSMPLEstimator',
    'MultiViewMultiPersonEnd2EndEstimator',
    'MultiViewMultiPersonTopDownEstimator',
    'MultiViewSinglePersonSMPLEstimator'
]



from mmcv.utils import Registry

from .smplify_base_hook import SMPLifyBaseHook
from .smplify_verbose_hook import SMPLifyVerboseHook

SMPLIFY_HOOKS = Registry('smplify_hook')
SMPLIFY_HOOKS.register_module(
    name='SMPLifyVerboseHook', module=SMPLifyVerboseHook)


def build_smplify_hook(cfg) -> SMPLifyBaseHook:
    """Build a hook for smplify."""
    return SMPLIFY_HOOKS.build(cfg)



import cv2
import matplotlib.pyplot as plt
import numpy as np
import os
import torch

from xrmocap.data_structure.keypoints import Keypoints
from xrmocap.transform.convention.keypoints_convention import get_keypoint_idx
from xrmocap.transform.limbs import get_limbs_from_keypoints

distribution = dict(
    mean=[
        0.29743698, 0.28764493, 0.86562234, 0.86257052, 0.31774172, 0.32603399,
        0.27688682, 0.28548218, 0.42981244, 0.43392589, 0.44601327, 0.43572195
    ],
    std=[
        0.02486281, 0.02611557, 0.07588978, 0.07094158, 0.04725651, 0.04132808,
        0.05556177, 0.06311393, 0.04445206, 0.04843436, 0.0510811, 0.04460523
    ],
    kps2conns={
        (0, 1): 0,
        (1, 0): 0,
        (0, 2): 1,
        (2, 0): 1,
        (0, 7): 2,
        (7, 0): 2,
        (0, 8): 3,
        (8, 0): 3,
        (1, 3): 4,
        (3, 1): 4,
        (2, 4): 5,
        (4, 2): 5,
        (3, 5): 6,
        (5, 3): 6,
        (4, 6): 7,
        (6, 4): 7,
        (7, 9): 8,
        (9, 7): 8,
        (8, 10): 9,
        (10, 8): 9,
        (9, 11): 10,
        (11, 9): 10,
        (10, 12): 11,
        (12, 10): 11
    })


def get_distance(x: np.ndarray, y: np.ndarray) -> np.float64:
    """Get 2d-2d point distance, return error if any of the input is nan.

    Args:
        x (np.ndarray): 2d point
        y (np.ndarray): 2d point

    Returns:
        np.float64: distance between two 2d points
    """
    if np.isnan(x).any() or np.isnan(y).any():
        return np.nan
    else:
        return ((x[0] - y[0])**2 + (x[1] - y[1])**2)**0.5


def projected_distance(pts_0,
                       pts_1,
                       F,
                       scores_0=None,
                       scores_1=None,
                       n_kps2d=17):
    """Compute point distance with epipolar geometry knowledge.

    :param pts_0: numpy points array with shape Nx17x2
    :param pts_1: numpy points array with shape Nx17x2
    :param F: Fundamental matrix F_{01}
    :return: numpy array of pairwise distance
    """

    lines = cv2.computeCorrespondEpilines(pts_0.reshape(-1, 1, 2), 2, F)
    lines = lines.reshape(-1, n_kps2d, 1, 3)
    lines = lines.transpose(0, 2, 1, 3)
    points_1 = np.ones([1, pts_1.shape[0], n_kps2d, 3])
    points_1[0, :, :, :2] = pts_1

    dist = np.sum(lines * points_1, axis=3)
    dist = np.abs(dist)
    dist = np.mean(dist, axis=2)

    return dist


def geometry_affinity(points_set: np.ndarray, Fs: np.ndarray,
                      dim_group: torch.Tensor, factor=5, n_kps2d=17) \
                      -> np.ndarray:
    """Get geometry affinity.

    Args:
        points_set (np.ndarray): Keypoints of the model, in shape
            (total number of people detected, n_kps2d, 2)
        Fs (np.ndarray): (num_cam, num_cam, 3, 3)
        dim_group (torch.Tensor): The cumulative number of people from
            different perspectives. e.g. If you have five cameras and
            each camera detects two people, dim_group is
            tensor([ 0,  2,  4,  6,  8, 10]).
        factor (int, optional): Defaults to 5.
        n_kps2d (int, optional): the number of keypoints2d. Defaults to 17.

    Returns:
        affinity_matrix (np.ndarray): affinity matrix in shape (N, N),
            N = n1+n2+..., n1 is the number of detected people in cam1
    """
    M, _, _ = points_set.shape
    dist = np.ones((M, M), dtype=np.float32) * factor**2
    np.fill_diagonal(dist, 0)
    # TODO: remove this stupid nested for loop
    for cam_id0, h in enumerate(range(len(dim_group) - 1)):
        for cam_add, k in enumerate(range(cam_id0 + 1, len(dim_group) - 1)):
            cam_id1 = cam_id0 + cam_add + 1
            # if there is no one in some view, skip it!
            if dim_group[h] == dim_group[h +
                                         1] or dim_group[k] == dim_group[k +
                                                                         1]:
                continue

            kps2d_id0 = points_set[dim_group[h]:dim_group[h + 1]]
            kps2d_id1 = points_set[dim_group[k]:dim_group[k + 1]]
            dist[dim_group[h]:dim_group[h+1], dim_group[k]:dim_group[k+1]]\
                = projected_distance(kps2d_id0, kps2d_id1,
                                     Fs[cam_id0, cam_id1],
                                     n_kps2d=n_kps2d)/2\
                + projected_distance(kps2d_id1, kps2d_id0,
                                     Fs[cam_id1, cam_id0],
                                     n_kps2d=n_kps2d).T/2
            dist[dim_group[k]:dim_group[k+1], dim_group[h]:dim_group[h+1]] =\
                dist[dim_group[h]:dim_group[h+1],
                     dim_group[k]:dim_group[k+1]].T
    if dist.std() < factor:
        for i in range(dist.shape[0]):
            dist[i, i] = dist.mean()

    affinity_matrix = -(dist - dist.mean()) / (dist.std() + 1e-12)
    # TODO: add flexible factor
    affinity_matrix = 1 / (1 + np.exp(-factor * affinity_matrix))
    return affinity_matrix


def check_bone_length(kps3d: np.ndarray, convention: str = 'coco') -> bool:
    """Check selected bone length.

    Args:
        kps3d (np.ndarray): 3xN 3D keypoints in MSCOCO order.
        convention (str, optional): Keypoints factory.. Defaults to 'coco'.

    Raises:
        NotImplementedError: The type of connection is not available。

    Returns:
        bool: If true, the selected bone length is satisfaction
    """
    min_length = 0.1
    max_length = 0.7
    kps3d = kps3d.transpose(1, 0)
    kps3d_score = np.ones((kps3d.shape[0], 1))
    kps3d = np.concatenate((kps3d, kps3d_score), axis=1)
    keypoints3d = Keypoints(kps=kps3d, convention=convention)
    limbs = get_limbs_from_keypoints(
        keypoints=keypoints3d, fill_limb_names=True)
    all_conn_dict = limbs.get_connections_by_names()
    selected_conn = []

    if convention == 'coco':
        selected_conn_name = [
            'left_lower_leg', 'left_thigh', 'right_lower_leg', 'right_thigh',
            'left_upperarm', 'right_upperarm', 'left_forearm', 'right_forearm'
        ]
        for key, value in all_conn_dict.items():
            if key in selected_conn_name:
                selected_conn.append(value)
    else:
        raise NotImplementedError('Other conventions are not yet implemented.')
    error_cnt = 0
    for kp_0, kp_1 in selected_conn:
        conn_length = np.sqrt(np.sum((kps3d[kp_0, :3] - kps3d[kp_1, :3])**2))
        if conn_length < min_length or conn_length > max_length:
            error_cnt += 1
    return error_cnt < 3


def visualize_match(frame_id,
                    n_camera,
                    matched_list,
                    sub_imgid2cam,
                    img_bboxes,
                    track_id,
                    input_folder,
                    img_folder,
                    cam_offset=4,
                    save_folder='./result'):
    data_name = input_folder.split('/')[-1]
    cols = len(matched_list) + 1
    rows = sub_imgid2cam.max() + 2
    cam_offset = 0
    if data_name == 'campus':
        imgs = [
            cv2.imread(f'{img_folder}/{data_name}/img/' +
                       f'Camera{cam_id}/campus4-c{cam_id}-{frame_id:05d}.png')
            for cam_id in range(cam_offset, cam_offset + n_camera)
        ]
    elif 'panoptic' in data_name:
        imgs = [
            cv2.imread(f'{img_folder}/{data_name}/img/' +
                       f'Camera{cam_id}/frame_{frame_id:06d}.png')
            for cam_id in range(cam_offset, cam_offset + n_camera)
        ]
    elif data_name == 'shelf':
        imgs = [
            cv2.imread(f'{img_folder}/{data_name}/img/' +
                       f'Camera{cam_id}/img_{frame_id:06d}.png')
            for cam_id in range(cam_offset, cam_offset + n_camera)
        ]
    else:
        NotImplementedError
    for i, person in enumerate(matched_list):
        # Plot origin image
        for sub_imageid in person:
            cam_id = sub_imgid2cam[sub_imageid]
            bbox = img_bboxes[sub_imageid]
            bbox[bbox < 0] = 0
            pid = track_id[sub_imageid]
            cropped_img = imgs[cam_id][int(bbox[1]):int(bbox[3]),
                                       int(bbox[0]):int(bbox[2])]

            plt.subplot(rows, cols, cam_id * cols + i + 2)
            cropped_img = cv2.cvtColor(cropped_img, cv2.COLOR_RGB2BGR)
            plt.imshow(cropped_img)
            plt.xlabel(f'#{sub_imageid}@{pid}')
            plt.xticks([])
            plt.yticks([])

    os.makedirs(f'{save_folder}', exist_ok=True)
    plt.savefig(f'{save_folder}/match_{int(frame_id)}.png')
    plt.close()


def get_min_reprojection_error(person, mview_proj_mat, kps2d_mat,
                               sub_imgid2cam):
    reproj_error = np.zeros((len(person), len(person)))
    for i, p0 in enumerate(person):
        for j, p1 in enumerate(person):
            projmat_0 = mview_proj_mat[sub_imgid2cam[p0]]
            projmat_1 = mview_proj_mat[sub_imgid2cam[p1]]
            kps2d_0, kps2d_1 = kps2d_mat[p0].T, kps2d_mat[p1].T
            kps3d_homo = cv2.triangulatePoints(projmat_0, projmat_1, kps2d_0,
                                               kps2d_1)
            this_error = 0
            for pk in person:
                projmat_k = mview_proj_mat[sub_imgid2cam[pk]]
                projected_kps_k_homo = projmat_k @ kps3d_homo
                projected_kps_k = \
                    projected_kps_k_homo[:2] / projected_kps_k_homo[2]
                this_error += np.linalg.norm(projected_kps_k - kps2d_mat[pk].T)
            reproj_error[i, j] = this_error

    reproj_error[np.arange(len(person)), np.arange(len(person))] = np.inf
    # TODO: figure out why NaN
    reproj_error[np.isnan(reproj_error)] = np.inf
    x, y = np.where(reproj_error == reproj_error.min())
    sub_imageid = np.array([person[x[0]], person[y[0]]])
    return sub_imageid


def check_limb_is_correct(model_start_point: np.ndarray,
                          model_end_point: np.ndarray,
                          gt_strat_point: np.ndarray,
                          gt_end_point: np.ndarray,
                          alpha=0.5) -> bool:
    """Check that limb predictions are correct.

    Returns:
        bool: If the predicted limb is correct, return True.
    """
    limb_lenth = np.linalg.norm(gt_end_point - gt_strat_point)
    start_difference = np.linalg.norm(gt_strat_point - model_start_point)
    end_difference = np.linalg.norm(gt_end_point - model_end_point)
    return ((start_difference + end_difference) / 2) <= alpha * limb_lenth


def vectorize_distance(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """Calculate euclid distance on each row of a and b.

    Args:
        a (np.ndarray): Points in shape [N, ...]
        b (np.ndarray): Points in shape [M, ...]

    Returns:
        np.ndarray: Dist in shape [MxN, ...]
        representing correspond distance.
    """
    N = a.shape[0]
    a = a.reshape(N, -1)
    M = b.shape[0]
    b = b.reshape(M, -1)
    a2 = np.tile(np.sum(a**2, axis=1).reshape(-1, 1), (1, M))
    b2 = np.tile(np.sum(b**2, axis=1), (N, 1))
    dist = a2 + b2 - 2 * (a @ b.T)
    dist = np.sqrt(dist)
    dist[np.where(np.isnan(dist))] = 1000
    return dist


def add_campus_jaw_headtop(nose, kps3d_campus):
    for frame_idx in range(nose.shape[0]):
        for i, kps3d in enumerate(kps3d_campus[frame_idx]):
            add_kps3d = np.zeros((2, 3))

            hip_center = (kps3d[2] + kps3d[3]) / 2
            shouler_cneter = (kps3d[8] + kps3d[9]) / 2
            head_center = nose[frame_idx][i]
            add_kps3d[0] = shouler_cneter + (head_center -
                                             shouler_cneter) * 0.5
            face_dir = np.cross(shouler_cneter - hip_center,
                                kps3d[8] - kps3d[9])
            if np.isnan(face_dir).any():
                add_kps3d[1] = shouler_cneter + (
                    head_center - shouler_cneter) * np.array([0.75, 0.75, 1.5])
            else:
                face_dir = face_dir / np.linalg.norm(face_dir)
                z_dir = np.array([0., 0., 1.], dtype=np.float)
                add_kps3d[1] = add_kps3d[0] + face_dir * 0.125 + z_dir * 0.145

            kps3d[-2:] = add_kps3d
    return kps3d_campus


def add_campus_jaw_headtop_mask(kps3d_mask):
    mask = np.zeros_like(kps3d_mask[:, :, -2:])
    for i, f_kps3d_mask in enumerate(kps3d_mask):
        kps3d_idx = np.sum(f_kps3d_mask, axis=1) > 0
        n_person = len(np.where(kps3d_idx)[0])
        mask[i][kps3d_idx] = np.ones((n_person, 2), dtype=np.uint8)
    ret_kps3d_mask = np.concatenate((kps3d_mask[:, :, :-2], mask), axis=-1)
    return ret_kps3d_mask


def align_by_keypoint(keypoints: Keypoints, keypoint_name='right_ankle'):
    convention = keypoints.get_convention()
    index = get_keypoint_idx(name=keypoint_name, convention=convention)
    if index == -1:
        raise ValueError('Check the convention of kps3d!')
    kps = keypoints.get_keypoints()[0, 0, :, :3]
    root = kps[index, :]
    return kps - root


def compute_mpjpe(pred_keypoints: Keypoints,
                  gt_keypoints: Keypoints,
                  align=False):
    """Compute MPJPE given prediction and ground-truth.

    Args:
        pred_keypoints (Keypoints)
        gt_keypoints (Keypoints)
        align (bool, optional): boolean value that determines whether to align
                                with the root. Defaults to False.

    Returns:
        MPJPE of the input keypoints.
    """
    if align:
        pred = align_by_keypoint(pred_keypoints, 'right_ankle')
        gt = align_by_keypoint(gt_keypoints, 'right_ankle')
    else:
        pred = pred_keypoints.get_keypoints()[..., :3]
        gt = gt_keypoints.get_keypoints()[..., :3]
    mpjpe = np.sqrt(np.sum(np.square(pred - gt), axis=-1))
    return mpjpe



# yapf: disable
import logging
import numpy as np
import torch
from typing import List, Union
from xrprimer.utils.log_utils import get_logger

# yapf: enable


class Limbs():
    """A class for person limbs data, recording connection vectors between
    keypoints.

    Connections are the only necessary data, while human parts, points are
    optional.
    """

    def __init__(self,
                 connections: Union[np.ndarray, torch.Tensor],
                 connection_names: Union[List[str], None] = None,
                 parts: Union[List[List[int]], None] = None,
                 part_names: Union[List[str], None] = None,
                 points: Union[np.ndarray, torch.Tensor, None] = None,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """A class for person limbs data, recording connection vectors between
        keypoints. Connections are the only necessary data. Connections record
        point indice, while parts record connection indice.

        Args:
            connections (Union[np.ndarray, torch.Tensor]):
                A tensor or ndarray for connections,
                in shape [n_conn, 2],
                conn[:, 0] are start point indice and
                conn[:, 1] are end point indice.
            connection_names (Union[List[str], None], optional):
                A list of connections names. If given,
                len(connection_names)==len(conn), else default names
                will be returned when getting connections.
                Defaults to None.
            parts (Union[List[List[int]], None], optional):
                A nested list, len(parts) is part number,
                and len(parts[0]) is connection number of the
                first part. Each element in parts[i] is an index
                of one connection.
            part_names (Union[List[str], None], optional):
                A list of part names. If given,
                len(part_names)==len(parts), else default names
                will be returned when getting parts.
                Defaults to None.
            points (Union[np.ndarray, torch.Tensor, None], optional):
                A tensor or ndarray for points,
                in shape [n_point, point_dim].
                Defaults to None.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        self.logger = get_logger(logger)
        self.connections = None
        self.connection_names = None
        self.parts = None
        self.part_names = None
        self.points = None
        self.logger = get_logger(None)
        self.set_connections(connections, connection_names)
        if parts is not None:
            self.set_parts(parts, part_names)
        if points is not None:
            self.set_points(points)

    def set_connections(self,
                        conn: Union[np.ndarray, torch.Tensor],
                        conn_names: List[str] = None) -> None:
        """Set connection relations of the limbs. Names are optional.

        Args:
            conn (Union[np.ndarray, torch.Tensor]):
                A tensor or ndarray for connections,
                in shape [n_conn, 2],
                conn[:, 0] are start point indice and
                conn[:, 1] are end point indice.
            conn_names (List[str], optional):
                A list of connections names. If given,
                len(conn_names)==len(conn), else default names
                will be returned when getting connections.
                Defaults to None.

        Raises:
            TypeError: Type of connections is not correct.
            ValueError: Shape of connections is not correct.
        """
        if isinstance(conn, torch.Tensor):
            connections = conn.detach().cpu().numpy()
        elif isinstance(conn, np.ndarray):
            connections = conn
        else:
            self.logger.error('Type of connections is not correct.\n' +
                              f'Type: {type(conn)}.')
            raise TypeError
        connections = connections.astype(dtype=np.int32)
        # shape: connection_number, 2
        if connections.shape[-1] != 2 or len(connections.shape) != 2:
            self.logger.error('Shape of connections should be' +
                              ' [n_conn, 2].\n' +
                              f'connections.shape: {connections.shape}.')
            raise ValueError
        self.connections = connections
        if conn_names is not None:
            if len(conn_names) == len(connections):
                self.connection_names = conn_names
            else:
                self.logger.warning(
                    'Length of connection_names is wrong, reset to None.\n' +
                    f'len(conn_names): {len(conn_names)}\n' +
                    f'len(connections): {len(connections)}')
                self.connection_names = None
        else:
            self.connection_names = None

    def set_parts(self,
                  parts: List[List[int]],
                  part_names: List[str] = None) -> None:
        """Set parts of the limbs. If parts has been set, connections can be
        arranged by part when getting. Names are optional.

        Args:
            parts (List[List[int]]):
                A nested list, len(parts) is part number,
                and len(parts[0]) is connection number of the
                first part. Each element in parts[i] is an index
                of one connection.
            part_names (List[str], optional):
                A list of part names. If given,
                len(part_names)==len(parts), else default names
                will be returned when getting parts.
                Defaults to None.

        Raises:
            TypeError: Type of parts is not correct.
            ValueError: Type of connection index is not correct.
        """
        if not isinstance(parts, list):
            self.logger.error('Type of parts is not correct.\n' +
                              f'Type: {type(parts)}.\n' + 'Expect: list')
            raise TypeError
        # Type of conn index: int
        for conn_list in parts:
            for conn_index in conn_list:
                if not isinstance(conn_index, int):
                    self.logger.error(
                        'Type of connection index is not correct.\n' +
                        f'Type: {type(conn_index)}.\n' + 'Expect: int')
                    raise TypeError
        self.parts = parts
        if part_names is not None:
            if len(part_names) == len(parts):
                self.part_names = part_names
            else:
                self.logger.warning(
                    'Length of part_names is wrong, reset to None.\n' +
                    f'len(part_names): {len(part_names)}\n' +
                    f'len(parts): {len(part_names)}')
                self.part_names = None
        else:
            self.part_names = None

    def set_points(self, points: Union[np.ndarray, torch.Tensor]) -> None:
        """Set points of the limbs.

        Args:
            points (Union[np.ndarray, torch.Tensor]):
                A tensor or ndarray for points,
                in shape [n_point, point_dim].

        Raises:
            TypeError:
                Type of points is not correct.
            ValueError:
                Shape of points should be [n_point, point_dim].
        """
        if isinstance(points, torch.Tensor):
            points = points.detach().cpu().numpy()
        elif not isinstance(points, np.ndarray):
            self.logger.error('Type of points is not correct.\n' +
                              f'Type: {type(points)}.')
            raise TypeError
        # shape: connection_number, 2
        if len(points.shape) != 2:
            self.logger.error('Shape of points should be' +
                              ' [n_point, point_dim].\n' +
                              f'points.shape: {points.shape}.')
            raise ValueError
        self.points = points

    def clone(self) -> 'Limbs':

        def copy_if_not_None(data):
            if data is None:
                return None
            else:
                return data.copy()

        ret_limbs = Limbs(
            connections=self.connections.copy(),
            connection_names=self.connection_names,
            parts=copy_if_not_None(self.parts),
            part_names=self.part_names,
            points=copy_if_not_None(self.points),
            logger=self.logger)
        return ret_limbs

    def get_points(self) -> Union[np.ndarray, None]:
        """Get points array, which might be None.

        Returns:
            Union[np.ndarray, None]: keypoints
        """
        return self.points

    def get_connections(self) -> np.ndarray:
        """Get connections of limbs.

        Returns:
            np.ndarray: connections
        """
        return self.connections

    def get_connection_names(self) -> List[str]:
        """Get names of connection.

        Returns:
            List[str]: A list of names
        """
        if self.connection_names is None:
            ret_list = [
                f'conn_{conn_index:03d}'
                for conn_index in range(len(self.connections))
            ]
        else:
            ret_list = self.connection_names
        return ret_list

    def get_parts(self) -> Union[List[List[int]], None]:
        """Get parts of limbs, which might be None.

        Returns:
            Union[List[List[int]], None]: parts
        """
        return self.parts

    def get_part_names(self) -> List[str]:
        """Get names of part. If self.part is None, an empty list will be
        returned.

        Returns:
            List[str]: A list of names
        """
        if self.parts is None:
            ret_list = []
        else:
            if self.part_names is None:
                ret_list = [
                    f'part_{part_index:03d}'
                    for part_index in range(len(self.parts))
                ]
            else:
                ret_list = self.part_names
        return ret_list

    def __len__(self) -> int:
        """Get number of connections.

        Returns:
            int
        """
        return len(self.connections)

    def get_connections_in_parts(self) -> dict:
        """Get connection in parts. In each part, there's a list of.

        [start_pt_index, end_pt_index].

        Returns:
            dict: keys are part names.
        """
        ret_dict = {}
        part_names = self.get_part_names()
        for part_index, part_name in enumerate(part_names):
            connection_list = self.parts[part_index]
            ret_dict[part_name] = connection_list
        return ret_dict

    def get_connections_by_names(self) -> dict:
        """Get connection by names.

        Returns:
            dict:
                keys are connection names and
                values are [start_pt_index, end_pt_index],
                in type ndarray.
        """
        ret_dict = {}
        conn_names = self.get_connection_names()
        for conn_index, conn_name in enumerate(conn_names):
            connection = self.connections[conn_index]
            ret_dict[conn_name] = connection
        return ret_dict



from .affinity_estimator import AppearanceAffinityEstimator
from .base_architecture import BaseArchitecture
from .multi_view_pose_transformer import MviewPoseTransformer

__all__ = [
    'AppearanceAffinityEstimator', 'BaseArchitecture', 'MviewPoseTransformer'
]



CAMPUS_KEYPOINTS = [
    'right_ankle', 'right_knee', 'right_hip_extra', 'left_hip_extra',
    'left_knee', 'left_ankle', 'right_wrist', 'right_elbow', 'right_shoulder',
    'left_shoulder', 'left_elbow', 'left_wrist', 'jaw', 'headtop'
]
